/*
 *	wiiuse
 *
 *	Written By:
 *		Michael Laforest	< para >
 *		Email: < thepara (--AT--) g m a i l [--DOT--] com >
 *
 *	Copyright 2005
 *
 *	This file is part of wiiuse.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	$Header$
 *
 */

/**
 *	@file
 *	@brief Handles wiimote events.
 *
 *	The file includes functions that handle the events
 *	that are sent from the wiimote to us.
 */

#include <stdio.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <malloc.h>

#include "definitions.h"
#include "connect.h"
#include "wiiuse_internal.h"
#include "dynamics.h"
#include "events.h"

static void propagate_event(struct wiimote_t* wm, byte event, byte* msg);
static void event_data_read(struct wiimote_t* wm, int* buttons, byte* msg);
static void event_status(struct wiimote_t* wm, int* buttons, byte* msg);


/**
 *	@brief Poll the wiimotes for any events.
 *
 *	@param wm		An array of pointers to wiimote_t structures.
 *	@param wiimotes	The number of wiimote_t structures in the \a wm array.
 *
 *	It is necessary to poll the wiimote devices for events
 *	that occur.  If an event occurs on a particular wiimote,
 *	the registered callback function is invoked.
 */
void wiimote_poll(struct wiimote_t** wm, int wiimotes) {
	struct timeval tv;
	fd_set fds;
	int r;
	int i;
	int highest_fd = 0;

	if (!wm) return;

	/* block select() for 1/20th of a second */
	tv.tv_sec = 0;
	tv.tv_usec = 50000;

	FD_ZERO(&fds);

	for (i = 0; i < wiimotes; ++i) {
		/* only poll it if it is connected */
		if (WIIMOTE_IS_SET(wm[i], WIIMOTE_STATE_CONNECTED)) {
			FD_SET(wm[i]->in_sock, &fds);

			/* find the highest fd of the connected wiimotes */
			if (wm[i]->in_sock > highest_fd)
				highest_fd = wm[i]->in_sock;
		}
	}

    if (select(highest_fd + 1, &fds, NULL, NULL, &tv) == -1) {
    	ERROR("Unable to select() the wiimote interrupt socket(s).");
    	return;
    }

	/* check each socket for an event */
	for (i = 0; i < wiimotes; ++i) {
		/* if this wiimote is not connected, skip it */
		if (!WIIMOTE_IS_SET(wm[i], WIIMOTE_STATE_CONNECTED))
			continue;

		if (!FD_ISSET(wm[i]->in_sock, &fds))
			/* no event */
			continue;

		/* clear out the event buffer */
		memset(wm[i]->event, 0, sizeof(wm[i]->event));

		/* read the pending message into the buffer */
		r = read(wm[i]->in_sock, wm[i]->event, sizeof(wm[i]->event));
		if (r == -1) {
			/* error reading data */
			ERROR("Receiving wiimote data (id %i).", wm[i]->unid);
			continue;
		}
		if (!r) {
			/* remote disconnect */
			wiimote_disconnected(wm[i]);
			continue;
		}

		/* if no event callback was set then we can't handle this */
		if (!wm[i]->event_cb)
			continue;

		/* TODO - what is 0xa1? */
		if (wm[i]->event[0] != 0xa1)  {
			WARNING("event[0] was not 0xa1 [is %x] -- skipping event (id %i)", wm[i]->event[0], wm[i]->unid);
			continue;
		}

		/* propagate the event */
		propagate_event(wm[i], wm[i]->event[1], wm[i]->event+2);
	}
}


/**
 *	@brief Analyze the event that occured on a wiimote.
 *
 *	@param wm		An array of pointers to wiimote_t structures.
 *	@param event	The event that occured.
 *	@param msg		The message specified in the event packet.
 *
 *	Pass the event to the registered event callback.
 */
static void propagate_event(struct wiimote_t* wm, byte event, byte* msg) {
	if (!wm)
		return;

	/* if no event callback was set then we can't handle this */
	if (!wm->event_cb)
		return;

	int buttons = 0;
	struct vec3b_t accel = {0};
	struct orient_t orient = {0};
	struct gforce_t gforce = {0};

	switch (event) {

		case WIIMOTE_REPORT_BTN:			/* payload = 2 */
		{
			/* button */
			buttons = wiimote_pressed_buttons(msg);
			break;
		}
		case WIIMOTE_REPORT_BTN_MOTION:		/* payload = 5 */
		{
			/* button - motion */
			buttons = wiimote_pressed_buttons(msg);

			accel.x = msg[2];
			accel.y = msg[3];
			accel.z = msg[4];

			/* calculate the remote orientation */
			calculate_orientation(wm, &accel, &orient);

			/* calculate the gforces on each axis */
			calculate_gforce(wm, &accel, &gforce);

			break;
		}
		case WIIMOTE_REPORT_READ:			/* payload = variable */
		{
			/* data read */
			event_data_read(wm, &buttons, msg);

			/* yeah buttons may be pressed, but this wasn't an "event" */
			return;
		}
		case WIIMOTE_REPORT_CTRL_STATUS:	/* payload = 6 */
		{
			/* controller status */
			event_status(wm, &buttons, msg);

			/* don't execute the event callback */
			return;
		}
		default:
		{
			WARNING("Unknown event, can not handle it [%x].", event);
			return;
		}
	}

	/* set the buttons released this event */
	wm->buttons_released = (wm->buttons_held & ~buttons);

	/* invoke the callback function */
	wm->event_cb(wm, buttons, accel, orient, gforce);

	/* set buttons currently pressed as being held */
	wm->buttons_held = buttons;
}


/**
 *	@brief Find what buttons are pressed.
 *
 *	@param msg		The message specified in the event packet.
 */
short wiimote_pressed_buttons(byte* msg) {
	short bits;

	/* convert to big endian */
	bits = BIG_ENDIAN_SHORT(*(short*)msg);

	return bits;
}


/**
 *	@brief Received a data packet from a read request.
 *
 *	@param wm		Pointer to a wiimote_t structure.
 *	@param buttons	[out] The buttons that were pressed during the event.
 *	@param msg		The message specified in the event packet.
 *
 *	Data from the wiimote comes in packets.  If the requested
 *	data segment size is bigger than one packet can hold then
 *	several packets will be received.  These packets are first
 *	reassembled into one, then the registered callback function
 *	that handles data reads is invoked.
 */
static void event_data_read(struct wiimote_t* wm, int* buttons, byte* msg) {
	/*
				btns? SE FF FF data
				vvvvv vv vv vv v------>
		(a1) 21 80 00 f0 11 f0 80 6c 8c c7 c2 5d 2e bd 40 00 4e 00 99 80 08 b1
	*/

	/* we must always assume the packet received is from the most recent request */
	byte err;
	byte len;
	unsigned short offset;
	struct read_req_t* req = wm->read_req;

	*buttons = wiimote_pressed_buttons(msg);

	/* if we don't have a request out then we didn't ask for this packet */
	if (!req) {
		WARNING("Received data packet when no request was made.");
		return;
	}

	err = msg[2] & 0x0F;

	if (err == 0x08) {
		WARNING("Unable to read data - address does not exist.");
		return;
	} else if (err == 0x07) {
		WARNING("Unable to read data - address is for write-only registers.");
		return;
	} else if (err) {
		WARNING("Unable to read data - unknown error code %x.", err);
		return;
	}

	len = ((msg[2] & 0xF0) >> 4) + 1;
	offset = BIG_ENDIAN_SHORT(*(unsigned short*)(msg + 3));

	req->wait -= len;
	if (req->wait >= req->size)
		/* this should never happen */
		req->wait = 0;

	DEBUG("Received read packet:");
	DEBUG("    Packet read offset:   %i bytes", offset);
	DEBUG("    Request read offset:  %i bytes", req->offset);
	DEBUG("    Read offset into buf: %i bytes", offset - req->offset);
	DEBUG("    Read data size:       %i bytes", len);
	DEBUG("    Still need:           %i bytes", req->wait);

	/* reconstruct this part of the data */
	memcpy((req->buf + offset - req->offset), (msg + 5), len);

	/* if all data has been received, execute the read event callback */
	if (!req->wait) {
		req->cb(wm, req->buf, req->size);

		/* delete this request */
		wm->read_req = req->next;
		free(req);

		/* if another request exists send it to the wiimote */
		if (wm->read_req)
			wiimote_send_next_pending_read_request(wm);
	}
}


/**
 *	@brief Read the controller status.
 *
 *	@param wm		Pointer to a wiimote_t structure.
 *	@param buttons	[out] The buttons that were pressed during the event.
 *	@param msg		The message specified in the event packet.
 *
 *	Read the controller status and execute the registered status callback.
 *
 *	At the time of writing this, bytes 3 and 4 of the message are of unknown purpose.
 */
static void event_status(struct wiimote_t* wm, int* buttons, byte* msg) {
	float battery_level;
	int led[4] = {0};
	int attachment = 0;
	int speaker = 0;
	int ir = 0;

	*buttons = wiimote_pressed_buttons(msg);

	printf(">>>> %x %x %x %x %x %x\n", msg[0], msg[1], msg[2], msg[3], msg[4], msg[5]);

	if (!wm->stat_cb)
		/* no status callback */
		return;

	/* find what LEDs are lit */
	if (msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_LED_1)	led[0] = 1;
	if (msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_LED_2)	led[1] = 1;
	if (msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_LED_3)	led[2] = 1;
	if (msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_LED_4)	led[3] = 1;

	/* is an attachment connected to the expansion port? */
	if ((msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_ATTACHMENT) == WIIMOTE_CTRL_STATUS_BYTE1_ATTACHMENT)
		attachment = 1;

	/* is the speaker enabled? */
	if ((msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_SPEAKER_ENABLED) == WIIMOTE_CTRL_STATUS_BYTE1_SPEAKER_ENABLED)
		speaker = 1;

	/* is IR sensing enabled? */
	if ((msg[2] & WIIMOTE_CTRL_STATUS_BYTE1_IR_ENABLED) == WIIMOTE_CTRL_STATUS_BYTE1_IR_ENABLED)
		ir = 1;

	/* find the battery level and normalize between 0 and 1 */
	battery_level = (msg[5] / (float)WIIMOTE_MAX_BATTERY_CODE);

	/* execute the callback */
	wm->stat_cb(wm, attachment, speaker, ir, led, battery_level);

	/*
	 *	From now on the remote will only receive status packets.
	 *	Why Nintendo did this I may never know, it seems pointless.
	 *	Perhaps it has something to do with the (at this time)
	 *	unknown expansion interface.
	 *	Anyhow, we need to send a WIIMOTE_CMD_REPORT_TYPE
	 *	packet to reenable other incomming reports.
	 */
	wiimote_send_report_type(wm);
}

